from __future__ import annotations
from dataclasses import dataclass
import logging as log
import requests

import knowledge_mapper.knowledge_interaction as knowledge_interaction

from knowledge_mapper.tke_exceptions import UnexpectedHttpResponseError

@dataclass(kw_only=True)
class KnowledgeBaseRegistrationRequest:
    """Class with necessary data itemsfor registering a new knowledge base."""
    id: str
    name: str
    description: str

class KnowledgeBase:
    def __init__(self, req: KnowledgeBaseRegistrationRequest, ke_url: str):
        self.ke_url = ke_url
        self.id = req.id
        self.name = req.name
        self.description = req.description
        self.kis = dict()
        self.kis_by_name = dict()


    def from_json(kb_json: dict, ke_url: str) -> KnowledgeBase:
        return KnowledgeBase(
            KnowledgeBaseRegistrationRequest(
                id=kb_json['knowledgeBaseId'],
                name=kb_json['knowledgeBaseName'],
                description=kb_json['knowledgeBaseDescription']
            ),
            ke_url=ke_url
        )


    def unregister(self):
        if self.ke_url is None:
            raise Exception("Cannot unregister this KB because no knowledge engine URL is known for this object.")

        response = requests.delete(
            f'{self.ke_url}/sc',
            headers={'Knowledge-Base-Id': self.id}
        )

        if not response.ok:
            raise UnexpectedHttpResponseError(response)


    def register_knowledge_interaction(self, ki: knowledge_interaction.KnowledgeInteractionRegistrationRequest, name=None) -> knowledge_interaction.KnowledgeInteraction:
        body = {
            'knowledgeInteractionType': ki.type,
            'prefixes': ki.prefixes
        }

        if name is not None:
            body['knowledgeInteractionName'] = name

        if isinstance(ki, knowledge_interaction.AskKnowledgeInteractionRegistrationRequest | knowledge_interaction.AnswerKnowledgeInteractionRegistrationRequest):
            body['graphPattern'] = ki.pattern
        elif isinstance(ki, knowledge_interaction.PostKnowledgeInteractionRegistrationRequest | knowledge_interaction.ReactKnowledgeInteractionRegistrationRequest):
            body['argumentGraphPattern'] = ki.argument_pattern
            body['resultGraphPattern'] = ki.result_pattern
        else:
            raise Exception('`ki` must be a concrete knowledge interaction object')

        response = requests.post(
            f'{self.ke_url}/sc/ki',
            headers={'Knowledge-Base-Id': self.id},
            json=body
        )

        if not response.ok:
            raise UnexpectedHttpResponseError(response)
        
        ki_id = response.json()['knowledgeInteractionId']

        log.info(f'Successfully registered knowledge interaction {ki_id}.')

        registered_ki = knowledge_interaction.KnowledgeInteraction.from_req(ki, ki_id, self)
        self.kis[ki_id] = registered_ki

        if name is not None:
            self.kis_by_name[name] = registered_ki

        return registered_ki


    def get_ki(self, name=None, id=None) -> knowledge_interaction.KnowledgeInteraction:
        if name is not None and id is None:
            if name in self.kis_by_name:
                return self.kis_by_name[name]
        if id is not None and name is None:
            if id in self.kis:
                return self.kis[id]
        return None


    def start_handle_loop(self, loops=None):
        did_loops = 0
        while loops is not None and did_loops < loops or loops is None:
            status, maybe_ke_request = self.long_poll()
            did_loops += 1
            if status == 'repoll':
                continue
            elif status == 'exit':
                log.warn('KE returned status code 410, meaning I have to exit')
                break
            elif status == 'handle':
                ke_request = maybe_ke_request

                # Extract relevant data items from the KE request
                ki_id = ke_request['knowledgeInteractionId']
                handle_request_id = ke_request['handleRequestId']
                bindings = ke_request['bindingSet']
                requesting_kb_id = ke_request['requestingKnowledgeBaseId']

                ki = self.kis[ki_id]
                if isinstance(ki, knowledge_interaction.AnswerKnowledgeInteraction):
                    result_bindings = ki.answer(bindings, requesting_kb_id)
                elif isinstance(ki, knowledge_interaction.ReactKnowledgeInteraction):
                    result_bindings = ki.react(bindings, requesting_kb_id)
                
                self.post_handle_response(ki, handle_request_id, result_bindings)


    def long_poll(self):
        log.info('Waiting for response to long poll...')
        response = requests.get(f'{self.ke_url}/sc/handle', headers = {'Knowledge-Base-Id': self.id})
        if response.status_code == 202:
            log.info('Received 202.')
            return "repoll", None
        elif response.status_code == 500:
            log.error(response.text)
            log.error('TKE had an internal server error. Reinitiating long poll.')
            return "repoll", None
        elif response.status_code == 410:
            return "exit", None
        elif response.status_code == 200:
            return "handle", response.json()
        else:
            log.warn(f'long_poll received unexpected status {response.status_code}')
            log.warn(response.text)
            log.warn('repolling anyway..')
            return "repoll", None


    def post_handle_response(self, ki, handle_id, bindings):
        response = requests.post(f'{self.ke_url}/sc/handle',
            json={
                'handleRequestId': handle_id,
                'bindingSet': bindings,
            },
            headers={
                'Knowledge-Base-Id': self.id,
                'Knowledge-Interaction-Id': ki.id,
            }
        )

        if not response.ok:
            raise UnexpectedHttpResponseError(response)



import argparse
import logging as log
import pyjson5
import sys
import importlib
import time
import signal

from knowledge_mapper.knowledge_mapper import KnowledgeMapper
from knowledge_mapper.auth.sql_auth import SqlAuth
from knowledge_mapper.auth.static_auth import StaticAuth
from knowledge_mapper.data_source import DataSource
from knowledge_mapper.sparql_source import SparqlSource
from knowledge_mapper.sql_source import SqlSource

log.basicConfig(level=log.INFO)

# This function is called when a SIGTERM signal is received. This makes it so
# that the knowledge mapper can be gracefully killed by Docker.
def handle_sigterm(*args):
    raise KeyboardInterrupt()

signal.signal(signal.SIGTERM, handle_sigterm)

DATA_SOURCE_MAX_CONNECTION_ATTEMPTS = 10
DATA_SOURCE_WAIT_BEFORE_RETRY = 3

def test_data_source(data_source: DataSource):
    success = False
    attempts = 0
    while not success:
        try:
            data_source.test()
            success = True
        except Exception as e:
            attempts += 1
            if attempts < DATA_SOURCE_MAX_CONNECTION_ATTEMPTS:
                log.warning(f'Request to data source failed. Retrying in {DATA_SOURCE_WAIT_BEFORE_RETRY} s.')
                time.sleep(DATA_SOURCE_WAIT_BEFORE_RETRY)
            else:
                log.error(f'Request to data source failed.')
                raise e


def main():
    from . import __version__
    log.info(f'Running Knowledge Mapper {__version__}')
    parser = argparse.ArgumentParser(description='Expose an endpoint to a knowledge network.')
    parser.add_argument('config')
    args = parser.parse_args()
    with open(args.config) as config_file:
        config = pyjson5.load(config_file)

        if 'sparql' in config:
            endpoint = config['sparql']['endpoint']
            username = None
            password = None
            if 'username_environment_var' in config['sparql'] and 'password_environment_var' in config['sparql']:
                username = config['sparql']['username_environment_var']
                password = config['sparql']['password_environment_var']
            data_source = SparqlSource(endpoint, username, password)
        elif 'sql_host' in config:
            data_source = SqlSource(config['sql_host'], config['sql_port'], config['sql_database'], config['sql_user'], config['sql_password'])
        elif 'plugin' in config:
            plugin_cfg = config['plugin']
            module_name, class_name = plugin_cfg['class'].rsplit(".", 1)
            plugin_module = importlib.import_module(module_name)
            plugin_class = getattr(plugin_module, class_name)
            if 'args' in plugin_cfg:
                data_source = plugin_class(*plugin_cfg['args'])
            else:
                data_source = plugin_class()
        else:
            log.error('Invalid config.')
            sys.exit(1)

        test_data_source(data_source)

        if 'authorization_enabled' in config:
            if 'authorization' in config:
                log.error('Cannot use both `authorization_enabled` and `authorization`, as `authorization_enabled=true` is a shorthand for `authorization={type="static"}`.')
                sys.exit(1)

            if config['authorization_enabled'] == True:
                auth_config = {'type': 'static'}
            elif config['authorization_enabled'] == False:
                auth_config = None
            else:
                log.error('"authorization_enabled" must be either "true" or "false"')
                sys.exit(1)
        elif 'authorization' in config:
            auth_config = config['authorization']
        else:
            auth_config = None

        if auth_config is not None:
            if auth_config['type'] == 'sql':
                authorization = SqlAuth(auth_config)
            elif auth_config['type'] == 'static':
                authorization = StaticAuth(auth_config)
            else:
                log.error('Unknown authorization type "%s"', auth_config['type'])
                sys.exit(1)
        else:
            authorization = None

        km = KnowledgeMapper(
            data_source,
            authorization,
            config['knowledge_engine_endpoint'],
            config['knowledge_base']['id'],
            config['knowledge_base']['name'],
            config['knowledge_base']['description']
        )
        for ki in config['knowledge_interactions']:
            km.add_knowledge_interaction(ki)

        try:
            km.start()
        except KeyboardInterrupt:
            log.info('Shutting down gracefully...')
        finally:
            km.clean_up()

        log.info('Goodbye.')
        exit(0)

if __name__ == '__main__':
    main()
